@@ -68,8 +68,10 @@ static void CairoGD_MetricInfo(int c,
 			      double* width, NewDevDesc *dd);
 static void CairoGD_Mode(int mode, NewDevDesc *dd);
 static void CairoGD_NewPage(R_GE_gcontext *gc, NewDevDesc *dd);
+#if R_GE_version >= 8
 static void CairoGD_Path(double *x, double *y, int npoly, int *nper, Rboolean winding,
                        R_GE_gcontext *gc, NewDevDesc *dd);
+#endif
 static void CairoGD_Polygon(int n, double *x, double *y,
 			   R_GE_gcontext *gc,
 			   NewDevDesc *dd);
@@ -96,19 +98,45 @@ static void CairoGD_TextUTF8(double x, double y, constxt char *str,
 			 double rot, double hadj,
 			 R_GE_gcontext *gc,
 			 NewDevDesc *dd);
+#if R_GE_version >= 6
 static void CairoGD_Raster(unsigned int *raster, int w, int h,
                        double x, double y, double width, double height,
                        double rot, Rboolean interpolate,
                        R_GE_gcontext *gc, NewDevDesc *dd);
+#endif
 static SEXP CairoGD_Cap(NewDevDesc *dd);
+#if R_GE_version >= 9
 static int  CairoGD_HoldFlush(NewDevDesc *dd, int level);
-/* additional GE 13 API */
+#endif
+
+#if R_GE_version >= 13 /* R 4.1.0 */
 static SEXP     CairoGD_setPattern(SEXP pattern, pDevDesc dd);
 static void     CairoGD_releasePattern(SEXP ref, pDevDesc dd);
 static SEXP     CairoGD_setClipPath(SEXP path, SEXP ref, pDevDesc dd);
 static void     CairoGD_releaseClipPath(SEXP ref, pDevDesc dd);
 static SEXP     CairoGD_setMask(SEXP path, SEXP ref, pDevDesc dd);
 static void     CairoGD_releaseMask(SEXP ref, pDevDesc dd);
+#define hasPatternFill(GC) (GC->patternFill != R_NilValue)
+#else
+#define hasPatternFill(GC) 0
+#endif
+
+/* GE 14 - R 4.1.0 (no fns, just deviceVersion and deviceClip) */
+
+#if R_GE_version >= 15 /* R 4.2.0 */
+/* this is required since 15 otherwise R segfaults */
+static SEXP CairoGD_Capabilities(SEXP capabilities);
+/* there are more below (groups, paths, lum. masks)
+   which we don't bother declaring here */
+#endif
+
+#if R_GE_version >= 16
+/* GE 16 glyph API (R 4.3.0) */
+static void CairoGD_Glyph(int n, int *glyphs, double *x, double *y,
+						  SEXP font, double size,
+						  int col, double rot, pDevDesc dd);
+#endif
+/* GE 17 = R 4.6.0 */
 
 /* fake mbcs support for old R versions */
 #if R_GE_version < 4
@@ -127,6 +155,50 @@ static void     CairoGD_releaseMask(SEXP ref, pDevDesc dd);
 
 int Rcairo_symbol_font_use_pua = 1;
 
+/* cairo 1.17+ has removed Win32 font support in cairo_select_font_face.
+   patch it back in if it has WIN32 font support */
+#if defined(WIN32) && CAIRO_VERSION > CAIRO_VERSION_ENCODE(1, 16, 0) && defined(CAIRO_HAS_WIN32_FONT)
+
+/* Select font face using win32 cairo backend instead of the default
+   which doesn't find all fonts */
+static void win32_cairo_select_font_face(cairo_t *cr, const char *family,
+										 cairo_font_slant_t slant, cairo_font_weight_t weight) {
+	LOGFONTW lf;
+	cairo_font_face_t *ff;
+
+	memset(&lf, 0, sizeof(LOGFONTW));
+	if (mbstowcs(lf.lfFaceName, family, LF_FACESIZE - 1) < 1) {
+		/* this is unlikely to work, but can try ... */
+		cairo_select_font_face(cr, family, slant, weight);
+		return;
+	}
+	lf.lfFaceName[LF_FACESIZE - 1] = L'\0';
+
+	lf.lfWeight = (weight == CAIRO_FONT_WEIGHT_BOLD) ? FW_BOLD : FW_NORMAL;
+	if (slant == CAIRO_FONT_SLANT_ITALIC ||
+		slant == CAIRO_FONT_SLANT_OBLIQUE)
+		lf.lfItalic = 1;
+	lf.lfCharSet = DEFAULT_CHARSET;
+	lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
+	lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
+	lf.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
+
+	ff = cairo_win32_font_face_create_for_logfontw(&lf);
+	if (ff && cairo_font_face_status(ff) == CAIRO_STATUS_SUCCESS) {
+		cairo_set_font_face(cr, ff);
+		cairo_font_face_destroy(ff);
+	} else /* fall back to default if it fails */
+		cairo_select_font_face(cr, family, slant, weight);
+}
+
+/* re-map cairo_select_font_face to win32_cairo_select_font_face */
+#ifdef cairo_select_font_face
+#undef cairo_select_font_face
+#endif
+#define cairo_select_font_face win32_cairo_select_font_face
+
+#endif /* Win32 fall-back */
+
 #if USE_CAIRO_FT
 FT_Library Rcairo_ft_library = NULL;
 
@@ -362,7 +434,7 @@ static void Rcairo_setup_font(CairoGDDesc* xd, R_GE_gcontext *gc) {
     Rprintf("INFO: set font[%d] size to %.f (cex=%.f, ps=%.f, fontscale=%.f)\n", i,
 			(gc->cex * gc->ps * xd->fontscale + 0.5) * 64, gc->cex, gc->ps, xd->fontscale);
 #endif
-	long new_size = (gc->cex * gc->ps * xd->fontscale + 0.5) * 64;
+	long new_size = (long) ((gc->cex * gc->ps * xd->fontscale + 0.5) * 64.0);
 	if (Rcairo_fonts[i].font_size != new_size || !Rcairo_fonts[i].hb_font) {
 #ifdef JGD_DEBUG
 		Rprintf("INFO: allocating new instance of font %d for size %ld\n", i, new_size);
@@ -377,7 +449,7 @@ static void Rcairo_setup_font(CairoGDDesc* xd, R_GE_gcontext *gc) {
 #endif
 }
 
-static void Rcairo_set_line(CairoGDDesc* xd, R_GE_gcontext *gc) {
+void Rcairo_set_line(CairoGDDesc* xd, R_GE_gcontext *gc) {
 	cairo_t *cc = xd->cb->cc;
 	cairo_line_cap_t lend = CAIRO_LINE_CAP_SQUARE;
 	cairo_line_join_t ljoin = CAIRO_LINE_JOIN_ROUND;
@@ -415,10 +487,32 @@ static void Rcairo_set_line(CairoGDDesc* xd, R_GE_gcontext *gc) {
 	}
 }
 
+#if R_GE_version >= 13
+/* from coreutil.c - replaces cairo_fill_preserve() but also supports pattern fills */
+void cairoFill(const R_GE_gcontext *gc, CairoGDDesc *xd);
+#else /* otherwise no pattern fills */
+void cairoFill(const R_GE_gcontext *gc, CairoGDDesc *xd) {
+	if (R_ALPHA(gc->fill) > 0) {
+		if(!xd || !xd->cb) return;
+		cairo_t *cc = xd->cb->cc;
+        cairo_set_antialias(cc, CAIRO_ANTIALIAS_NONE);
+		Rcairo_set_color(cc, gc->col);
+		cairo_fill_preserve(cc);
+        cairo_set_antialias(cc, xd->antialias);
+    }
+}
+#endif
+
 /*------- the R callbacks begin here ... ------------------------*/
 
 #define c_trunc(X) ((double)((int)(X)))  /* some Suns don't have trunc(), so we use the cast-way to truncate */
 
+/* from coreutil.c */
+#define ct_bool int
+#define ct_false 0
+ct_bool cairoBegin(CairoGDDesc *xd);
+void cairoEnd(ct_bool grouping, CairoGDDesc *xd);
+
 static void CairoGD_Activate(NewDevDesc *dd)
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
@@ -480,25 +574,27 @@ static void CairoGD_Circle(double x, double y, double r,  R_GE_gcontext *gc,  Ne
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
 	if(!xd || !xd->cb) return;
-	{
-		cairo_t *cc = xd->cb->cc;
+	cairo_t *cc = xd->cb->cc;
 
 #ifdef JGD_DEBUG
-		Rprintf("circle %x %f/%f %f [%08x/%08x]\n",cc, x, y, r, gc->col, gc->fill);
+	Rprintf("circle %x %f/%f %f [%08x/%08x]\n",cc, x, y, r, gc->col, gc->fill);
 #endif
-
+	if (xd->appending) {
+		cairo_new_sub_path(cc);
+		cairo_arc(cc, x, y, r + 0.5 , 0., 2 * M_PI); /* Add 0.5 like devX11 */
+	} else {
+		ct_bool grouping = cairoBegin(xd);
 		cairo_new_path(cc);
 		cairo_arc(cc, x, y, r + 0.5 , 0., 2 * M_PI); /* Add 0.5 like devX11 */
-		if (CALPHA(gc->fill)) {
-			Rcairo_set_color(cc, gc->fill);
-			cairo_fill_preserve(cc);
-		}
+		if (hasPatternFill(gc) || CALPHA(gc->fill))
+			cairoFill(gc, xd);
 		if (CALPHA(gc->col) && gc->lty!=-1) {
 			Rcairo_set_color(cc, gc->col);
 			Rcairo_set_line(xd, gc);
 			cairo_stroke(cc);
 		} else cairo_new_path(cc);
-		xd->cb->serial++;
+		cairoEnd(grouping, xd);
+		if (!grouping) xd->cb->serial++;
 	}
 }
 
@@ -522,6 +618,9 @@ static void CairoGD_Clip(double x0, double x1, double y0, double y1,  NewDevDesc
 #endif
 }
 
+/* coreutil.c */
+void CairoGD_CoreUtil_Destroy(CairoGDDesc *xd);
+
 static void CairoGD_Close(NewDevDesc *dd)
 {
   CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
@@ -537,6 +636,7 @@ static void CairoGD_Close(NewDevDesc *dd)
 	  R_ReleaseObject(xd->cb->onSave);
 	  xd->cb->onSave = 0;
   }
+  CairoGD_CoreUtil_Destroy(xd);
   xd->cb->destroy_backend(xd->cb);
 
   free(xd);
@@ -550,6 +650,7 @@ static void CairoGD_Deactivate(NewDevDesc *dd)
 	if (xd->cb->activation) xd->cb->activation(xd->cb, 0);
 }
 
+#if R_GE_version >= 9
 static int  CairoGD_HoldFlush(NewDevDesc *dd, int level)
 {
 	int ol;
@@ -569,6 +670,7 @@ static int  CairoGD_HoldFlush(NewDevDesc *dd, int level)
     }
     return xd->holdlevel;
 }
+#endif
 
 static Rboolean CairoGD_Locator(double *x, double *y, NewDevDesc *dd)
 {
@@ -582,27 +684,33 @@ static void CairoGD_Line(double x1, double y1, double x2, double y2,  R_GE_gcont
 {
     CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
     if(!xd || !xd->cb) return;
-    
+	cairo_t *cc = xd->cb->cc;
+
 #ifdef JGD_DEBUG
     Rprintf("line %f/%f %f/%f [%08x/%08x]\n", x1, y1, x2, y2, gc->col, gc->fill);
 #endif
-
-    if (CALPHA(gc->col) && gc->lty!=-1) {
-      cairo_t *cc = xd->cb->cc;
-      cairo_new_path(cc);
-	  if ((x1==x2 || y1==y2) &&xd->cb->truncate_rect) {
-		  /* if we are snapping rectangles to grid, we also need to snap straight
-			 lines to make sure they match - e.g. tickmarks, baselines etc. */
-		  x1=c_trunc(x1)+0.5; x2=c_trunc(x2)+0.5;
-		  y1=c_trunc(y1)+0.5; y2=c_trunc(y2)+0.5;
-	  }
-      cairo_move_to(cc, x1, y1);
-      cairo_line_to(cc, x2, y2);
-      Rcairo_set_color(cc, gc->col);
-      Rcairo_set_line(xd, gc);
-      cairo_stroke(cc);
-	  xd->cb->serial++;
-    }
+	if (xd->appending) {
+		cairo_move_to(cc, x1, y1);
+		cairo_line_to(cc, x2, y2);
+		return;
+	}
+	if (CALPHA(gc->col) && gc->lty != -1) {
+		ct_bool grouping = cairoBegin(xd);
+		cairo_new_path(cc);
+		if ((x1 == x2 || y1 == y2) && xd->cb->truncate_rect) {
+			/* if we are snapping rectangles to grid, we also need to snap straight
+			   lines to make sure they match - e.g. tickmarks, baselines etc. */
+			x1 = c_trunc(x1) + 0.5; x2 = c_trunc(x2) + 0.5;
+			y1 = c_trunc(y1) + 0.5; y2 = c_trunc(y2) + 0.5;
+		}
+		cairo_move_to(cc, x1, y1);
+		cairo_line_to(cc, x2, y2);
+		Rcairo_set_color(cc, gc->col);
+		Rcairo_set_line(xd, gc);
+		cairo_stroke(cc);
+		cairoEnd(grouping, xd);
+		if (!grouping) xd->cb->serial++;
+	}
 }
 
 #ifdef HAVE_HARFBUZZ
@@ -617,7 +725,7 @@ static int str2utf16(const char *c, int len, UChar **buf, const char *ifrom) {
         return 0;
     }
     size_t osize = sizeof(UChar) * (ce - c + 1), isize = ce - c;
-    UChar *js = buf[0] = (osize < sizeof(js_buf)) ? js_buf : (UChar*) R_alloc(sizeof(UChar), ce - c + 1);
+    UChar *js = buf[0] = (osize < sizeof(js_buf)) ? js_buf : (UChar*) R_alloc(sizeof(UChar), (int) (ce - c + 1));
     char *dst = (char*) js;
     int end_test = 1, is_le = (((char*)&end_test)[0] == 1) ? 1 : 0;
     if (!ifrom) ifrom = "";
@@ -762,7 +870,7 @@ static rc_text_shape *c_setup_ascii_glyphs(CairoGDDesc *xd, R_GE_gcontext *gc, c
 	int i = xd->fontface - 1;
 	if (i < 0 || i > 8) i = 0;
 	Rcairo_font_face *rf = &Rcairo_fonts[i];
-	chb_add_glyphs(ts, rf, (const UChar*) str, 0, strlen(str), CHB_DIR_LTR | CHB_FIRST | CHB_LAST | CHB_UTF8);
+	chb_add_glyphs(ts, rf, (const UChar*) str, 0, (int32_t) strlen(str), CHB_DIR_LTR | CHB_FIRST | CHB_LAST | CHB_UTF8);
 	return ts;
 }
 
@@ -780,7 +888,7 @@ static rc_text_shape *c_setup_glyphs(CairoGDDesc *xd, R_GE_gcontext *gc, const c
 	int32_t ulen = 0;
 	if (!bidi) bidi = ubidi_open();
 	if (!bidi) Rf_error("Unable to allocate memory for UBiDi");
-	ulen = str2utf16(str, strlen(str), &text, encoding) / sizeof(UChar); /* str2utf16 returns bytes, need chars */
+	ulen = str2utf16(str, (int) strlen(str), &text, encoding) / sizeof(UChar); /* str2utf16 returns bytes, need chars */
 	ubidi_setPara(bidi, text, ulen, UBIDI_DEFAULT_LTR, NULL, &err);
 	if (U_FAILURE(err))
 		Rf_error("Unable to compute UBiDi for string '%s'", str);
@@ -855,7 +963,7 @@ static void CairoGD_MetricInfo(int c,  R_GE_gcontext *gc,  double* ascent, doubl
 		}
 #endif
 	} else {
-		str[0] = c; str[1] = 0;
+		str[0] = (char) c; str[1] = 0;
 		/* Here, we assume that c < 256 */
 	}
 
@@ -936,6 +1044,9 @@ static SEXP findArg(const char *name, SEXP list) {
 	return 0;
 }
 
+/* coreutil.c (patterns etc.) */
+void CairoGD_CoreUtil_Init(CairoGDDesc *xd);
+
 Rboolean CairoGD_Open(NewDevDesc *dd, CairoGDDesc *xd,  const char *type, int conn, const char *file, double w, double h,
 					  double umpl, SEXP aux)
 {
@@ -1059,6 +1170,8 @@ Rboolean CairoGD_Open(NewDevDesc *dd, CairoGDDesc *xd,  const char *type, int co
 	if (xd->dpix>0 && xd->dpiy>0) xd->asp = xd->dpix / xd->dpiy;
 
 	Rcairo_backend_init_surface(xd->cb);
+	CairoGD_CoreUtil_Init(xd);
+
 	/*
 	  cc = xd->cb->cc;
 	  cairo_save(cc);
@@ -1071,57 +1184,68 @@ Rboolean CairoGD_Open(NewDevDesc *dd, CairoGDDesc *xd,  const char *type, int co
 	return TRUE;
 }
 
+#if R_GE_version >= 9
+static void cc_define_path(cairo_t *cc, double *x, double *y, int npoly,
+                           int *nper, Rboolean winding) {
+	int i, j, n = 0;
+	for (i = 0; i < npoly; i++) {
+		cairo_move_to(cc, x[n], y[n]);
+		n++;
+		for(j = 1; j < nper[i]; j++) {
+			cairo_line_to(cc, x[n], y[n]);
+			n++;
+		}
+		cairo_close_path(cc);
+	}
+}
+
 static void CairoGD_Path(double *x, double *y, int npoly, int *nper, Rboolean winding,
                        R_GE_gcontext *gc, NewDevDesc *dd)
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
 	if(!xd || !xd->cb || !nper || npoly < 1) return;
-	{
-		int i, j, n;
-		cairo_t *cc = xd->cb->cc;
-		
+	cairo_t *cc = xd->cb->cc;
+#ifdef JGD_DEBUG
+	Rprintf("path %d polygons [%08x/%08x]\n", npoly, gc->col, gc->fill);
+#endif
+
+	if (xd->appending) {
+		cc_define_path(cc, x, y, npoly, nper, winding);
+	} else {
+		ct_bool grouping = cairoBegin(xd);
 		Rcairo_set_line(xd, gc);
 
-#ifdef JGD_DEBUG
-		Rprintf("path %d polygons [%08x/%08x]\n", npoly, gc->col, gc->fill);
-#endif
-		
 		cairo_new_path(cc);
-		n = 0;
-		for (i = 0; i < npoly; i++) {
-			cairo_move_to(cc, x[n], y[n]);
-			n++;
-			for(j = 1; j < nper[i]; j++) {
-				cairo_line_to(cc, x[n], y[n]);
-				n++;
-			}
-			cairo_close_path(cc);
-		}
-
-		if (CALPHA(gc->fill)) {
+		cc_define_path(cc, x, y, npoly, nper, winding);
+		if (hasPatternFill(gc) || CALPHA(gc->fill)) {
 			if (winding) 
 				cairo_set_fill_rule(cc, CAIRO_FILL_RULE_WINDING);
 			else 
 				cairo_set_fill_rule(cc, CAIRO_FILL_RULE_EVEN_ODD);
-			Rcairo_set_color(cc, gc->fill);
-			cairo_fill_preserve(cc);
+			cairoFill(gc, xd);
 		}
 		if (CALPHA(gc->col) && gc->lty != -1) {
 			Rcairo_set_color(cc, gc->col);
 			cairo_stroke(cc);
 		} else cairo_new_path(cc);
-		xd->cb->serial++;
+		cairoEnd(grouping, xd);
+		if (!grouping) xd->cb->serial++;
     }
 }
+#endif
 
 static void CairoGD_Polygon(int n, double *x, double *y,  R_GE_gcontext *gc,  NewDevDesc *dd)
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
 	if(!xd || !xd->cb || n<2) return;
-	{
-		int i=1;
-		cairo_t *cc = xd->cb->cc;
-
+	cairo_t *cc = xd->cb->cc;
+	int i = 0;
+	if (xd->appending) {
+		cairo_move_to(cc, x[0], y[0]);
+		while (++i < n) cairo_line_to(cc, x[i], y[i]);
+		cairo_close_path(cc);
+	} else {
+		ct_bool grouping = cairoBegin(xd);
 		Rcairo_set_line(xd, gc);
 
 #ifdef JGD_DEBUG
@@ -1129,17 +1253,16 @@ static void CairoGD_Polygon(int n, double *x, double *y,  R_GE_gcontext *gc,  Ne
 #endif
 		cairo_new_path(cc);
 		cairo_move_to(cc, x[0], y[0]);
-		while (i<n) { cairo_line_to(cc, x[i], y[i]); i++; }
+		while (++i < n) cairo_line_to(cc, x[i], y[i]);
 		cairo_close_path(cc);
-		if (CALPHA(gc->fill)) {
-			Rcairo_set_color(cc, gc->fill);
-			cairo_fill_preserve(cc);
-		}
+		if (hasPatternFill(gc) || CALPHA(gc->fill))
+			cairoFill(gc, xd);
 		if (CALPHA(gc->col) && gc->lty!=-1) {
 			Rcairo_set_color(cc, gc->col);
 			cairo_stroke(cc);
 		} else cairo_new_path(cc);
-		xd->cb->serial++;
+		cairoEnd(grouping, xd);
+		if (!grouping) xd->cb->serial++;
 	}
 }
 
@@ -1147,21 +1270,25 @@ static void CairoGD_Polyline(int n, double *x, double *y,  R_GE_gcontext *gc,  N
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
 	if(!xd || !xd->cb || n<2) return;
-	{
-		int i=1;
-		cairo_t *cc = xd->cb->cc;
-
+	cairo_t *cc = xd->cb->cc;
+	int i = 0;
+	if (xd->appending) {
+		cairo_move_to(cc, x[0], y[0]);
+		while (++i < n) cairo_line_to(cc, x[i], y[i]);
+	} else {
 #ifdef JGD_DEBUG
 		Rprintf("poly-line %d points [%08x]\n", n, gc->col);
 #endif
 		if (CALPHA(gc->col) && gc->lty!=-1) {
+			ct_bool grouping = cairoBegin(xd);
 			cairo_new_path(cc);
 			cairo_move_to(cc, x[0], y[0]);
-			while (i<n) { cairo_line_to(cc, x[i], y[i]); i++; }
+			while (++i < n) cairo_line_to(cc, x[i], y[i]);
 			Rcairo_set_color(cc, gc->col);
 			Rcairo_set_line(xd, gc);
 			cairo_stroke(cc);
-			xd->cb->serial++;
+			cairoEnd(grouping, xd);
+			if (!grouping) xd->cb->serial++;
 		}
 	}
 }
@@ -1170,13 +1297,17 @@ static void CairoGD_Rect(double x0, double y0, double x1, double y1,  R_GE_gcont
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
 	if(!xd || !xd->cb) return;
-	{
-		cairo_t *cc = xd->cb->cc;
+	cairo_t *cc = xd->cb->cc;
+
+	if (xd->appending) {
+		cairo_rectangle(cc, x0, y0, x1 - x0, y1 - y0);
+	} else {
 		double snap_add = 0.0;
 		if (x1<x0) { double h=x1; x1=x0; x0=h; }
 		if (y1<y0) { double h=y1; y1=y0; y0=h; }
 		/* if (x0<0) x0=0; if (y0<0) y0=0; */
 
+		ct_bool grouping = cairoBegin(xd);
 		Rcairo_set_line(xd, gc);
 
 #ifdef JGD_DEBUG
@@ -1196,10 +1327,8 @@ static void CairoGD_Rect(double x0, double y0, double x1, double y1,  R_GE_gcont
 
 		cairo_new_path(cc);
 		cairo_rectangle(cc, x0, y0, x1-x0+snap_add, y1-y0+snap_add);
-		if (CALPHA(gc->fill)) {
-			Rcairo_set_color(cc, gc->fill);
-			cairo_fill_preserve(cc);
-		}
+		if (hasPatternFill(gc) || CALPHA(gc->fill))
+			cairoFill(gc, xd);
 		if (CALPHA(gc->col) && gc->lty!=-1) {
 			/* if we are snapping, note that lines must be in 0.5 offset to fills in order
 			   hit the same pixels, because lines extend to both sides */
@@ -1210,17 +1339,19 @@ static void CairoGD_Rect(double x0, double y0, double x1, double y1,  R_GE_gcont
 			Rcairo_set_color(cc, gc->col);
 			cairo_stroke(cc);
 		} else cairo_new_path(cc);
-		xd->cb->serial++;
+		cairoEnd(grouping, xd);
+		if (!grouping) xd->cb->serial++;
 	}
 }
 
+#if R_GE_version >= 6
 static void CairoGD_Raster(unsigned int *raster, int w, int h,
                        double x, double y, double width, double height,
                        double rot, Rboolean interpolate,
                        R_GE_gcontext *gc, NewDevDesc *dd)
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
-	if(!xd || !xd->cb) return;
+	if(!xd || !xd->cb || xd->appending) return;
 	{   /* the code in this block has been adapted from R 2.12.1, src/modules/X11/cairoX11.c
 		   (c) 2010 R Development Core Team under GPL 2+ */
 		cairo_t *cc = xd->cb->cc;
@@ -1229,6 +1360,7 @@ static void CairoGD_Raster(unsigned int *raster, int w, int h,
 		int i;
 
 		cairo_save(cc);
+        ct_bool grouping = cairoBegin(xd);
 		cairo_translate(cc, x, y);
 		if (rot != 0.0)
 			cairo_rotate(cc, -rot * M_PI/180);
@@ -1244,12 +1376,12 @@ static void CairoGD_Raster(unsigned int *raster, int w, int h,
 		/* allocate data and transform to pre-mpl alpha */
 		imageData = (unsigned char *) malloc(4*w*h);
 		for (i = 0; i < w * h; i++) {
-			int alpha = R_ALPHA(raster[i]);
-			imageData[i*4 + 3] = alpha;
+			unsigned int alpha = (unsigned int) R_ALPHA(raster[i]);
+			imageData[i*4 + 3] = (unsigned char) alpha;
 			if (alpha < 255) {
-				imageData[i*4 + 2] = R_RED(raster[i]) * alpha / 255;
-				imageData[i*4 + 1] = R_GREEN(raster[i]) * alpha / 255;
-				imageData[i*4 + 0] = R_BLUE(raster[i]) * alpha / 255;
+				imageData[i*4 + 2] = (unsigned char) (R_RED(raster[i]) * alpha / 255);
+				imageData[i*4 + 1] = (unsigned char) (R_GREEN(raster[i]) * alpha / 255);
+				imageData[i*4 + 0] = (unsigned char) (R_BLUE(raster[i]) * alpha / 255);
 			} else {
 				imageData[i*4 + 2] = R_RED(raster[i]);
 				imageData[i*4 + 1] = R_GREEN(raster[i]);
@@ -1272,13 +1404,15 @@ static void CairoGD_Raster(unsigned int *raster, int w, int h,
 		cairo_rectangle(cc, 0, 0, w, h);
 		cairo_clip(cc);
 		cairo_paint(cc); 
-		
+        cairoEnd(grouping, xd);
+
 		cairo_restore(cc);
 		cairo_surface_destroy(image);
 		free(imageData);
-		xd->cb->serial++;
+		if (!grouping) xd->cb->serial++;
 	}
 }
+#endif
 
 static void CairoGD_Size(double *left, double *right,  double *bottom, double *top,  NewDevDesc *dd)
 {
@@ -1296,9 +1430,9 @@ static void CairoGD_Size(double *left, double *right,  double *bottom, double *t
 static double CairoGD_StrWidthEnc(constxt char *str,  R_GE_gcontext *gc,  NewDevDesc *dd, const char *encoding)
 {
 	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
-	int slen = strlen(str);
+	size_t slen = strlen(str);
 	if (!str) return 0;
-	if(!xd || !xd->cb) return slen*8;
+	if(!xd || !xd->cb) return ((double)slen) * 8.0;
 
 	Rcairo_setup_font(xd, gc);
 
@@ -1357,6 +1491,7 @@ static void CairoGD_TextEnc(double x, double y, constxt char *str,  double rot,
 #endif
 
 	cairo_save(cc);
+	ct_bool grouping = xd->appending ? ct_false : cairoBegin(xd);
 
 	cairo_translate(cc, x, y);
 
@@ -1373,7 +1508,7 @@ static void CairoGD_TextEnc(double x, double y, constxt char *str,  double rot,
 			cairo_translate(cc, -te.x_advance*hadj, 0);
 		/* Rcairo_set_color(cc, 0xff80ff); */
 	}
-	Rcairo_set_color(cc, gc->col);
+	if (!xd->appending) Rcairo_set_color(cc, gc->col);
 
 #ifdef HAVE_HARFBUZZ
 	cairo_show_glyphs(cc, ts->glyph, ts->glyphs);
@@ -1381,7 +1516,9 @@ static void CairoGD_TextEnc(double x, double y, constxt char *str,  double rot,
 	cairo_move_to(cc, 0, 0);
 	cairo_show_text(cc, str);
 #endif
-	xd->cb->serial++;
+	if (!xd->appending)
+		cairoEnd(grouping, xd);
+	if (!grouping) xd->cb->serial++;
 
 #ifdef JGD_DEBUG
 	{
@@ -1419,6 +1556,92 @@ static void CairoGD_TextNative(double x, double y, constxt char *str,
 	CairoGD_TextEnc(x, y, str, rot, hadj, gc, dd, "");
 }
 
+#if R_GE_version >= 15
+static SEXP CairoGD_Capabilities(SEXP capabilities) {
+#if R_GE_version >= 16
+	SEXP glyphs = PROTECT(ScalarInteger(1));
+	SET_VECTOR_ELT(capabilities, R_GE_capability_glyphs, glyphs);
+	UNPROTECT(1);
+#endif
+	return capabilities;
+}
+#endif
+
+#if R_GE_version >= 16
+static void CairoGD_Glyph(int n, int *glyphs, double *x, double *y,
+						  SEXP font, double size,
+						  int col, double rot, pDevDesc dd) {
+	CairoGDDesc *xd = (CairoGDDesc *) dd->deviceSpecific;
+	cairo_t *cc;
+
+	if (!xd || !xd->cb) return;
+
+	cc = xd->cb->cc;
+
+	ct_bool grouping = xd->appending ? ct_false : cairoBegin(xd);
+
+	/* font info is explicit so we don't use our regular cache and mapping */
+	double weight = R_GE_glyphFontWeight(font);
+	int style = R_GE_glyphFontStyle(font);
+	int wt = (weight > 400) ? CAIRO_FONT_WEIGHT_BOLD : CAIRO_FONT_WEIGHT_NORMAL;
+	int sl = (style == R_GE_text_style_normal) ? CAIRO_FONT_SLANT_NORMAL : CAIRO_FONT_SLANT_ITALIC;
+
+#ifdef JGD_DEBUG
+	Rprintf("INFO: Glyph(n=%d), fontfile='%s', index=%d, weight=%f, style=%d, col=%x, rot=%f\n",
+			n, R_GE_glyphFontFile(font) ? R_GE_glyphFontFile(font) : "<NULL>",
+			R_GE_glyphFontIndex(font), weight, style, col, rot);
+#endif
+
+	cairo_font_face_t *cairo_face = NULL;
+#if USE_CAIRO_FT
+	FcPattern *pattern = FcPatternBuild(NULL,
+										FC_FILE, FcTypeString,
+										R_GE_glyphFontFile(font),
+										FC_INDEX, FcTypeInteger,
+										R_GE_glyphFontIndex(font),
+										NULL);
+	cairo_face = cairo_ft_font_face_create_for_pattern(pattern);
+	FcPatternDestroy(pattern);
+#endif
+	if (cairo_face &&
+		cairo_font_face_status(cairo_face) == CAIRO_STATUS_SUCCESS) {
+		cairo_set_font_face(cc, cairo_face);
+	} else /* fall back - hopefully not needed */
+		cairo_select_font_face(cc, R_GE_glyphFontFamily(font), sl, wt);
+
+	/* FIXME: not sure if we need to scale - Quartz uses 72 baseline - check sizes */
+	cairo_set_font_size (cc, size / (72*dd->ipr[0]));
+
+	if (!xd->appending) Rcairo_set_color(cc, col);
+
+	int i = 0;
+	while (i < n) {
+		if (rot != 0.0) {
+			cairo_save(cc);
+			cairo_translate(cc, x[i], y[i]);
+			cairo_rotate(cc, -rot/180.*M_PI);
+			cairo_translate(cc, -x[i], -y[i]);
+		}
+
+		cairo_glyph_t cairoGlyph;
+		cairoGlyph.index = glyphs[i];
+		cairoGlyph.x = x[i];
+		cairoGlyph.y = y[i];
+
+		cairo_show_glyphs(cc, &cairoGlyph, 1);
+		/* for path ops use cairo_glyph_path(cc, &cairoGlyph, 1); instead */
+
+		if (rot != 0.0)
+			cairo_restore(cc);
+		i++;
+	}
+	if (!xd->appending)
+		cairoEnd(grouping, xd);
+	if (!grouping) xd->cb->serial++;
+}
+#endif
+
+#if R_GE_version >= 13
 static SEXP CairoGD_setPattern(SEXP pattern, pDevDesc dd) {
     return R_NilValue;
 }
@@ -1437,6 +1660,25 @@ static SEXP CairoGD_setMask(SEXP path, SEXP ref, pDevDesc dd) {
 }
 
 static void CairoGD_releaseMask(SEXP ref, pDevDesc dd) {}
+#endif
+
+#if R_GE_version >= 15
+static SEXP CairoGD_defineGroup(SEXP source, int op, SEXP destination, pDevDesc dd) {
+	return R_NilValue;
+}
+
+static void CairoGD_useGroup(SEXP ref, SEXP trans, pDevDesc dd) {}
+
+static void CairoGD_releaseGroup(SEXP ref, pDevDesc dd) {}
+
+static void CairoGD_stroke(SEXP path, const pGEcontext gc, pDevDesc dd) {}
+
+static void CairoGD_fill(SEXP path, int rule, const pGEcontext gc, pDevDesc dd) {}
+static void CairoGD_fillStroke(SEXP path, int rule, const pGEcontext gc, pDevDesc dd) {}
+#endif
+
+/** from coreutil.c **/
+void CairoGD_CoreUtil_SetupFn(NewDevDesc *dd);
 
 /** fill the R device structure with callback functions */
 void Rcairo_setup_gd_functions(NewDevDesc *dd) {
@@ -1475,11 +1717,26 @@ void Rcairo_setup_gd_functions(NewDevDesc *dd) {
     dd->releaseClipPath = CairoGD_releaseClipPath;
     dd->setMask         = CairoGD_setMask;
     dd->releaseMask     = CairoGD_releaseMask;
+/* 14 = deviceClip */
+#if R_GE_version >= 15
+	dd->defineGroup     = CairoGD_defineGroup;
+	dd->useGroup        = CairoGD_useGroup;
+	dd->releaseGroup    = CairoGD_releaseGroup;
+	dd->stroke          = CairoGD_stroke;
+	dd->fill            = CairoGD_fill;
+	dd->fillStroke      = CairoGD_fillStroke;
+	dd->capabilities    = CairoGD_Capabilities;
+#if R_GE_version >= 16
+	/* NOTE: caps must be set from 14 on or else R segfaults */
+	dd->glyph           = CairoGD_Glyph;
 #endif
 #endif
 #endif
 #endif
 #endif
+#endif
+#endif
+	CairoGD_CoreUtil_SetupFn(dd);
 }
 
 void Rcairo_backend_resize(Rcairo_backend *be, double width, double height) {
